﻿--[[
TODO:
  Bug: ghost on login tries to create new note!!

]]
---------------------------------------------------------
-- Addon declaration
HandyNotes_Charon = LibStub("AceAddon-3.0"):NewAddon("HandyNotes_Charon","AceEvent-3.0", "AceTimer-3.0")
local HC = HandyNotes_Charon
local Astrolabe = DongleStub("Astrolabe-0.4")
local L = LibStub("AceLocale-3.0"):GetLocale("HandyNotes_Charon")
local G = {}

---------------------------------------------------------
-- Our db upvalue and db defaults
local CURRENT_DB_VERSION = 1
local db
local defaults = {
	profile = {
		icon_scale = 1.0,
		icon_alpha = 1.0,
		filter = { ["*"] = true },
	},
	factionrealm = {
		dbversion = 0,
		nodes = {
			["*"] = {},
		}
	},
	char = {
		dbversion = 0,
		nodes = {
			["*"] = {},
		}
	},
}


---------------------------------------------------------
-- Localize some globals
local next = next
local select = select
local string_find = string.find
local GameTooltip = GameTooltip
local WorldMapTooltip = WorldMapTooltip
local HandyNotes = HandyNotes


---------------------------------------------------------
-- Constants

local defkey = {}
local iconDB = {
	["M"] = "Interface\\Minimap\\Tracking\\BattleMaster",       -- Melee Death
	["R"] = "Interface\\Minimap\\Tracking\\BattleMaster",       -- Range Death
	["S"] = "Interface\\Minimap\\Tracking\\BattleMaster",       -- Spell Death
	["E"] = "Interface\\Minimap\\Tracking\\BattleMaster",       -- Environment Death

	["H"] = "Interface\\Minimap\\Tracking\\Banker", -- SpiritHealer

	-- Default
	[defkey] = "Interface\\Minimap\\Tracking\\TrivialQuests", -- for DB errors??
		}

setmetatable(iconDB, {__index = function (t, k)
					local v = t[defkey]
					rawset(t, k, v) -- cache the value for next retrievals
					return v
				end})

local EnvirnmentsTable = {
	["DROWNING"] = "Drowned",
	["FALLING"] = "Falled to death",
	["FATIGUE"] = "Drowned from fatigue",
	["FIRE"] = "Burnt",
	["LAVA"] = "Burnt in lava",
	["SLIME"] = "Slimed",
}

---------------------------------------------------------
-- Plugin Handlers to HandyNotes

local HCHandler = {}

local function deletePin(button, mapFile, coord, type)
	local x, y = HandyNotes:getXY(coord)
	db[type].nodes[mapFile][coord] = nil
	HC:SendMessage("HandyNotes_NotifyUpdate", "Charon")
end
local function createWaypoint(button, mapFile, coord, type)
	local c, z = HandyNotes:GetCZ(mapFile)
	local x, y = HandyNotes:getXY(coord)
	local vType, vName, vGuild = strsplit(":", db[type].nodes[mapFile][coord])
	if TomTom then
		TomTom:AddZWaypoint(c, z, x*100, y*100, vName)
	elseif Cartographer_Waypoints then
		Cartographer_Waypoints:AddWaypoint(NotePoint:new(HandyNotes:GetCZToZone(c, z), x, y, vName))
	end
end

local clickedNoteCoord, clickedNoteZone, clickedNoteType

local info = {}
local function generateMenu(button, level)
	if (not level) then return end
	for k in pairs(info) do info[k] = nil end
	if (level == 1) then
		-- Create the title of the menu
		info.isTitle      = 1
		info.text         = L["HandyNotes - Charon"]
		info.notCheckable = 1
		UIDropDownMenu_AddButton(info, level)

		if TomTom or Cartographer_Waypoints then
			-- Waypoint menu item
			info.disabled     = nil
			info.isTitle      = nil
			info.notCheckable = nil
			info.text = L["Create waypoint"]
			info.icon = nil
			info.func = createWaypoint
			info.arg1 = clickedNoteZone
			info.arg2 = clickedNoteCoord
			info.arg3 = clickedNoteType
			UIDropDownMenu_AddButton(info, level);
		end

		-- Delete menu item
		info.disabled     = nil
		info.isTitle      = nil
		info.notCheckable = nil
		info.text = L["Delete note"]
		info.icon = nil
		info.func = deletePin
		info.arg1 = clickedNoteZone
		info.arg2 = clickedNoteCoord
		info.arg3 = clickedNoteType
		UIDropDownMenu_AddButton(info, level);

		-- Close menu item
		info.text         = L["Close"]
		info.icon         = nil
		info.func         = function() CloseDropDownMenus() end
		info.arg1         = nil
		info.arg2         = nil
		info.notCheckable = 1
		UIDropDownMenu_AddButton(info, level);
	end
end

local HC_Dropdown = CreateFrame("Frame", "HandyNotes_CharonDropdownMenu")
HC_Dropdown.displayMode = "MENU"
HC_Dropdown.initialize = generateMenu

function HCHandler:OnClick(button, down, mapFile, coord)
	if button == "RightButton" and not down then
		if db.char.nodes[mapFile][coord] then
			clickedNoteType = char
		elseif db.factionrealm.nodes[mapFile][coord] then
			clickedNoteType = factionrealm
		else
			return
		end
		clickedNoteZone = mapFile
		clickedNoteCoord = coord
		ToggleDropDownMenu(1, nil, HC_Dropdown, self, 0, 0)
	end
end


---------------------------------------------------------
-- Line drawing helper functions

-- Function to get the intersection point of 2 lines (x1,y1)-(x2,y2) and (sx,sy)-(ex,ey)
-- If there is no intersection point, it returns (x2, y2)
local function GetIntersection(x1, y1, x2, y2, sx, sy, ex, ey)
	local dx = x2-x1
	local dy = y2-y1
	local numer = dx*(sy-y1) - dy*(sx-x1)
	local demon = dx*(sy-ey) + dy*(ex-sx)
	if demon ~= 0 and dx ~= 0 then
		local u = numer / demon
		local t = (sx + (ex-sx)*u - x1)/dx
		if u >= 0 and u <= 1 and t >= 0 and t <= 1 then
			return sx + (ex-sx)*u, sy + (ey-sy)*u --return true
		end
	end
	return x2, y2 --return false
end

-- Function to draw a line between 2 coordinates on map (C,Z)
-- (x1,y1) is already translated to map (C,Z)
local function drawline(C, Z, x1, y1, mapFile2, coord2, color)
--[[	color = tonumber(color)
	if color == playerFaction then -- Same faction
		color = 1
	elseif color + playerFaction == 3 then -- Different faction
		if not db.show_both_factions then return end
		color = 2
	elseif color + playerFaction == 6 then -- Different faction, but special
		if not db.show_both_factions then return end
		color = 4
	end
	local C2, Z2 = HandyNotes:GetCZ(mapFile2)]]
	local x2, y2 = HandyNotes:getXY(coord2)
--	x2, y2 = Astrolabe:TranslateWorldMapPosition(C2, Z2, x2, y2, C, Z)
	x2, y2 = GetIntersection(x1, y1, x2, y2, 0, 0, 0, 1)
	x2, y2 = GetIntersection(x1, y1, x2, y2, 0, 0, 1, 0)
	x2, y2 = GetIntersection(x1, y1, x2, y2, 0, 1, 1, 1)
	x2, y2 = GetIntersection(x1, y1, x2, y2, 1, 0, 1, 1)
	local w, h = WorldMapButton:GetWidth(), WorldMapButton:GetHeight()
	if (x1 == x2) and (y1 == y2) then 
		return 
	end
	G:DrawLine(WorldMapButton, x1*w, (1-y1)*h, x2*w, (1-y2)*h, 25, color, "OVERLAY")
end


function HCHandler:OnEnter(mapFile, coord)
	local tooltip = self:GetParent() == WorldMapButton and WorldMapTooltip or GameTooltip
	if ( self:GetCenter() > UIParent:GetCenter() ) then -- compare X coordinate
		tooltip:SetOwner(self, "ANCHOR_LEFT")
	else
		tooltip:SetOwner(self, "ANCHOR_RIGHT")
	end
	local data
	if db.char.nodes[mapFile][coord] then
		data = db.char.nodes[mapFile][coord]
	elseif db.factionrealm.nodes[mapFile][coord] then
		data = db.factionrealm.nodes[mapFile][coord]
	else
		return
	end

	local vType, vTime, vMonster, vSPCoord = strsplit(":", data)
	local vName = "Error"
	local vDate = ""

	if vType == "H" then
		vName = "Spirit Healer"
	elseif vType == "M" then
		vName = "Melee Death"
		vDate = date("%c", vTime)
	elseif vType == "R" then
		vName = "Range Death"
		vDate = date("%c", vTime)
	elseif vType == "S" then
		vName = "Spell Death"
		vDate = date("%c", vTime)
	elseif vType == "E" then
		vName = "Environment Death"
		vDate = date("%c", vTime)
		if vMonster then
			vMonster = EnvirnmentsTable[vMonster]
		end
	end

	tooltip:AddLine("|cffe0e0e0"..vName.."|r")
	if vDate and (vDate ~= "") then tooltip:AddLine(vDate) end
	if vMonster and (vMonster ~= "") then tooltip:AddLine(vMonster) end
	if self:GetParent() == WorldMapButton then
		if vSPCoord and (vSPCoord ~= "") then 
			local x1, y1 = HandyNotes:getXY(coord)
			local x2, y2 = HandyNotes:getXY(vSPCoord)
			tooltip:AddLine(format("%.1f, %.1f", x2 * 100, y2 * 100))

			local C, Z = GetCurrentMapContinent(), GetCurrentMapZone()
			drawline(C, Z, x1, y1, mapFile, vSPCoord, {0, 0.5, 0, 1})
		end
	end

--	tooltip:AddLine(L["Charon"])
	tooltip:Show()
end

function HCHandler:OnLeave(mapFile, coord)
	if self:GetParent() == WorldMapButton then
		WorldMapTooltip:Hide()
		G:HideLines(WorldMapButton)
	else
		GameTooltip:Hide()
	end
end

do
--[[
	-- This is a custom iterator we use to iterate over every node in a given zone
	local function iter(t, prestate)
		if not t then return nil end
		local state, value = next(t, prestate)
		while state do
			if value then
				local vType = strsplit(":", value)
				if db.profile.filter[vType] then
					local icon = iconDB[vType]
					local scale = db.profile.icon_scale
					if vType ~= "H" then -- deaths smaller
						scale = scale * 0.75
					end
					return state, nil, icon, scale, db.profile.icon_alpha
				end
			end
			state, value = next(t, state)
		end
		return nil, nil, nil, nil
	end
	function HCHandler:GetNodes(mapFile)
		return iter, db.factionrealm.nodes[mapFile], nil
	end]]

	-- This is a custom iterator we use to iterate over every node in a given zone
	local function multiiter(lparam, oldcoord)
		if not lparam then return nil end

		--local idx

		if not lparam.current then
			--print("first table")
			lparam.current = next(lparam.tables, lparam.current)
		end

		--local t = lparam.current

		local coord, value = oldcoord, nil
		while lparam.current do
			local t = lparam.tables[lparam.current]
			coord, value = next(t, coord)
			while coord do
				if value then
					local vType = strsplit(":", value)
					if db.profile.filter[vType] then
						local icon = iconDB[vType]
						local scale = db.profile.icon_scale
						if vType ~= "H" then -- deaths smaller
							scale = scale * 0.75
						end
						return coord, nil, icon, scale, db.profile.icon_alpha
					end
				end
				coord, value = next(t, coord)
			end
			--print("nexting table")
			lparam.current = next(lparam.tables, lparam.current)
			--t = lparam.current
		end
		return nil, nil, nil, nil
	end
	function HCHandler:GetNodes(mapFile)
		--print("GetNodes()")
		local lparam = {}
		lparam.tables = {db.factionrealm.nodes[mapFile], db.char.nodes[mapFile]}
		lparam.current = nil
		return multiiter, lparam, nil
	end
end


local function GetVFilters()
	local vnds = {
	["D"] = L["TYPE_Auctioneer"],
	["S"] = L["TYPE_Banker"],
	["W"] = L["TYPE_BattleMaster"],
	["Z"] = L["TYPE_StableMaster"],
	["X"] = L["TYPE_SpiritHealer"],
		}

	local res = {}
	for id, text in pairs(vnds) do
		res[id] = "|T"..iconDB[id]..":18|t "..text
	end
	return res
end

---------------------------------------------------------
-- Options table

local options = {
	type = "group",
	name = "Charon",
	desc = "Charon",
	get = function(info) return db.profile[info.arg] end,
	set = function(info, v)
		db.profile[info.arg] = v
		HC:SendMessage("HandyNotes_NotifyUpdate", "Charon")
	end,
	args = {
		desc = {
			name = L["These settings control the look and feel of the Charon icons."],
			type = "description",
			order = 0,
		},
		icon_scale = {
			type = "range",
			name = L["Icon Scale"],
			desc = L["The scale of the icons"],
			min = 0.25, max = 2, step = 0.01,
			arg = "icon_scale",
			order = 10,
		},
		icon_alpha = {
			type = "range",
			name = L["Icon Alpha"],
			desc = L["The alpha transparency of the icons"],
			min = 0, max = 1, step = 0.01,
			arg = "icon_alpha",
			order = 20,
		},
--[[		filters = {
			type = "multiselect",
			name = L["Filters"],
			desc = nil,
			order = 30,
			width = "full",
			get = function(info, k) return db.profile.filter[k] end,
			set = function(info, k, v)
				db.profile.filter[k] = v
				HC:SendMessage("HandyNotes_NotifyUpdate", "Charon")
			end,
			values = GetVFilters(),

		}, ]]
	},
}


---------------------------------------------------------
-- NPC info tracking - TT handling

local tt = CreateFrame("GameTooltip")
tt:SetOwner(UIParent, "ANCHOR_NONE")
tt.left = {}
tt.right = {}

for i = 1, 30 do
	tt.left[i] = tt:CreateFontString()
	tt.left[i]:SetFontObject(GameFontNormal)
	tt.right[i] = tt:CreateFontString()
	tt.right[i]:SetFontObject(GameFontNormal)
	tt:AddFontStrings(tt.left[i], tt.right[i])
end


local LEVEL_start = "^" .. (type(LEVEL) == "string" and LEVEL or "Level")
local function FigureNPCGuild(unit)
	tt:ClearLines()
	tt:SetUnit(unit)
	if not tt:IsOwned(UIParent) then
		tt:SetOwner(UIParent, "ANCHOR_NONE")
	end
	local left_2 = tt.left[2]:GetText()
	if not left_2 or left_2:find(LEVEL_start) then
		return ""
	end
	return left_2
end

---------------------------------------------------------
-- Addon initialization, enabling and disabling


local thres = 25 -- in yards
local LastTime = 0;
local time = time;

local CorpseContinent = -1
local CorpseZoneId = -1
local CorpseX, CorpseY = 0, 0
local CorpseCoords = -1
local CorpseFound = false

local SpiritContinent = -1
local SpiritZoneId = -1
local SpiritX, SpiritY = 0, 0
local SpiritCoords = -1
local SpiritFound = false

local PlayerGuid

local LastDamageSource = ""
local LastDamageSourceType = "U"

function HC:OnInitialize()
	-- Set up our database
	db = LibStub("AceDB-3.0"):New("HandyNotes_CharonDB", defaults)
	self.db = db


	if (db.factionrealm.dbversion > CURRENT_DB_VERSION) or (db.char.dbversion > CURRENT_DB_VERSION) then
		print("|cff6fafffHandyNotes_Charon:|r |cffff4f00Warning:|r Unknown database version. Please update to newer version.")
		print("|cff6fafffHandyNotes_Charon:|r |cffff4f00Warning:|r Addon has been disabled to protect your database.")
		self:Disable()
		return
	end

	if db.factionrealm.dbversion ~= CURRENT_DB_VERSION then
		if db.factionrealm.dbversion == 0 then
			-- addon was just installed
			db.factionrealm.dbversion = CURRENT_DB_VERSION
		end
	end

	if db.char.dbversion ~= CURRENT_DB_VERSION then
		if db.char.dbversion == 0 then
			-- addon was just installed
			db.char.dbversion = CURRENT_DB_VERSION
		end
	end


	db.factionrealm.dbversion = CURRENT_DB_VERSION
	db.char.dbversion = CURRENT_DB_VERSION

	-- Initialize our database with HandyNotes
	HandyNotes:RegisterPluginDB("Charon", HCHandler, options)
end



function HC:OnEnable()

	PlayerGuid = UnitGUID("player")

	self:RegisterEvent("AREA_SPIRIT_HEALER_IN_RANGE")
	self:RegisterEvent("AREA_SPIRIT_HEALER_OUT_OF_RANGE")
	self:RegisterEvent("CONFIRM_XP_LOSS")
	self:RegisterEvent("CORPSE_IN_INSTANCE")
--	self:RegisterEvent("CORPSE_IN_RANGE")
--	self:RegisterEvent("CORPSE_OUT_OF_RANGE")
	self:RegisterEvent("RESURRECT_REQUEST")
	self:RegisterEvent("PLAYER_DEAD")
	self:RegisterEvent("PLAYER_ALIVE")
	self:RegisterEvent("PLAYER_UNGHOST")

	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")

--	self:RegisterAllEvents();
end


function HC:OnDisable()
	G:HideLines(WorldMapButton)
end


function HC:print(text)
	if DEFAULT_CHAT_FRAME then
		DEFAULT_CHAT_FRAME:AddMessage("|cffffff7fCharon: |r"..text)
	end
end

function HC:COMBAT_LOG_EVENT_UNFILTERED(event, ...)
	local type = select(2, ...)
	local dest = select(6, ...)

	local color = "0000ff"

	if (dest == PlayerGuid) then
		color = "ff0000"
	else
		return;
	end

	local overkillId = -1

	if (type == "SWING_DAMAGE") then
		overkillId = 8 + 0 + 2
		LastDamageSource = select(4, ...)
		LastDamageSourceType = "M"
	elseif (type == "RANGE_DAMAGE") then
		overkillId = 8 + 3 + 2
		LastDamageSource = select(4, ...)
		LastDamageSourceType = "R"
	elseif (type == "SPELL_DAMAGE") or (type == "SPELL_PERIODIC_DAMAGE") then
		overkillId = 8 + 3 + 2
		LastDamageSource = select(4, ...)
		LastDamageSourceType = "S"
	elseif (type == "ENVIRONMENTAL_DAMAGE") then
		overkillId = 8 + 1 + 2
		LastDamageSource = select(9, ...)
		LastDamageSourceType = "E"
	else
		return;
	end

--[[

	local txt = "|cff".. color .. type .. "|r: "
	for i = 3, select("#", ...) do
		txt = txt .. " " .. i .. ": " .. (select(i, ...) or "nil")
	end
	self:print(txt) ]]

--	self:print("1: "..select(1, ...).." 2: "..select(2, ...).." 3: "..(select(3, ...) or "nil").." 4: "..(select(4, ...) or "nil").." 5: "..(select(5, ...) or "nil").." 6: "..(select(6, ...) or "nil").." 7: "..(select(7, ...) or "nil").." 8: "..(select(8, ...) or "nil"))
--[[	if (type == "PARTY_KILL") then
		self:print(type)
		for i = 1, select("#", ...) do
			print("["..i.."] " .. (select(i, ...) or "nil"))
		end
	elseif (type == "UNIT_DIED") then
		self:print(type)
		for i = 1, select("#", ...) do
			print("["..i.."] " .. (select(i, ...) or "nil"))
		end
	end ]]

end

function HC:OnUpdate(event, ...)
	if (time() - LastTime > 1) then
		self:print("Update: " .. (event or "nil"))
	end
end


function HC:OnEvent(event, ...)
	if (time() - LastTime > 1) then
		self:print("Event: " .. (event or "nil"))
	end
--	self:print("Event: " .. (event or "nil"))
end

function HC:CONFIRM_XP_LOSS()
	self:print("CONFIRM_XP_LOSS")
end

function HC:AREA_SPIRIT_HEALER_IN_RANGE()
	self:print("AREA_SPIRIT_HEALER_IN_RANGE")
end

function HC:AREA_SPIRIT_HEALER_OUT_OF_RANGE()
	self:print("AREA_SPIRIT_HEALER_OUT_OF_RANGE")
end

function HC:CORPSE_IN_INSTANCE()
	self:print("CORPSE_IN_INSTANCE")
end

function HC:CORPSE_IN_RANGE()
	self:print("CORPSE_IN_RANGE")
end

function HC:CORPSE_OUT_OF_RANGE()
	self:print("CORPSE_OUT_OF_RANGE")
end

function HC:RESURRECT_REQUEST()
	self:print("RESURRECT_REQUEST")
end

function HC:PLAYER_DEAD()
	self:print("PLAYER_DEAD")
	if UnitIsDeadOrGhost("player") and select(2, IsInInstance()) ~= "pvp" and not IsActiveBattlefieldArena() then	
		self:print("We are dead.")
		self:RegisterCorpse()
		self:AddCorpseNote()
	end
end



function HC:FindAndUpdateCorpse()
	self:FindCorpse()
	if CorpseFound then
		self:AddOrUpdateCorpseNote()
	end
end

function HC:TryRegisterSpiritHealerAndUpdateCorpse()
	self:RegisterSpiritHealer()
	if SpiritFound then
		self:AddSpiritHealerNote()
		self:FindAndUpdateCorpse()
	end
end

function HC:PLAYER_ALIVE()
	self:print("PLAYER_ALIVE")
	if UnitIsDeadOrGhost("player") and select(2, IsInInstance()) ~= "pvp" and not IsActiveBattlefieldArena() then
		self:print("We are dead!")
		self:TryRegisterSpiritHealerAndUpdateCorpse()
	end
end

function HC:TestCorpseAdd()
	if UnitIsDeadOrGhost("player") and select(2, IsInInstance()) ~= "pvp" and not IsActiveBattlefieldArena() then	
		self:print("We are dead.")
		self:RegisterCorpse()
		self:AddCorpseNote()
	end

end

function HC:PLAYER_UNGHOST()
	self:print("PLAYER_UNGHOST")
end



function HC:RegisterCorpse()

	local Ccontinent, Czone, Cx, Cy = Astrolabe:GetCurrentPlayerPosition()
	if not Ccontinent then
		return
	end

	CorpseContinent = Ccontinent
	CorpseZoneId = Czone
	CorpseX, CorpseY = Cx, Cy
	CorpseCoords = HandyNotes:getCoord(Cx, Cy)
end

function HC:RegisterSpiritHealer()

	local Scontinent, Szone, Sx, Sy = Astrolabe:GetCurrentPlayerPosition()
	if not Scontinent then
		return
	end

	local Scoord = HandyNotes:getCoord(Sx, Sy)

	if Scoord == CorpseCoords then
		-- location not yet updated, need to retry after a while
		self:print("Spirit Healer not found, reschedulling")
		self:ScheduleTimer("TryRegisterSpiritHealerAndUpdateCorpse", 0.25)
		SpiritFound = false
		return
	else
		local Cx, Cy = CorpseX, CorpseY
		local dist = Astrolabe:ComputeDistance(Scontinent, Szone, Sx, Sy, Scontinent, Szone, Cx, Cy)
		if dist < 5 then
			-- location not yet updated, need to retry after a while
			self:print("Spirit Healer not found, reschedulling")
			self:ScheduleTimer("TryRegisterSpiritHealerAndUpdateCorpse", 0.25)
			SpiritFound = false
			return
		end
	end

	SpiritContinent = Scontinent
	SpiritZoneId = Szone
	SpiritX, SpiritY = Sx, Sy
	SpiritCoords = Scoord
	SpiritFound = true
end

-- This is a horribly convoluted methodology of finding your corpse cross zonally
-- Basically, since we can't depend on Cart to send locations like we
-- do for GuildPoint, we scan the map for the corpse.
-- 1. Locate the continent on which it's on
-- 2. Cycle through the zones til we find it
-- 3. Use the zone and local (to them) map coords to send to Cartographer.
-- The Cartographer map puts all waypoints into suspension on show so as not to lock
-- the map, since we're playing around with it so much to get locations.

function HC:FindCorpse()

--	self:print("Reading location")

		local Pcontinent, Pzone, Px, Py = Astrolabe:GetCurrentPlayerPosition()
--[[		if Pcontinent then
			self:print("Player Position: " .. Pcontinent .. ", " .. Pzone .. " [" .. Px .. ":" .. Py .. "]")
		else 
			self:print("Player Position: unknown")
		end ]]

	local x, y = 0, 0

	if CorpseContinent ~= -1 then -- If we have a cached continent we are on

--		self:print("Testing old continent: " .. CorpseContinent)

		SetMapZoom(CorpseContinent)
		x, y = GetCorpseMapPosition() -- See if we're still on that continent
	end

	if x == 0 and y == 0 then -- We're no longer on this continent or we never knew which we were on

--		self:print("Trying to find continent")
		CorpseContinent = -1

		for i=1,select("#",GetMapContinents()) do
			SetMapZoom(i) -- Cycle through the continents
			x, y = GetCorpseMapPosition()
			if x ~= 0 and y ~= 0 then
				CorpseContinent = i
				break
			end
		end
	end

	if (CorpseContinent == -1) then
		-- still no corpse, need to retry after a while
		self:print("corpse not found, reschedulling")
		self:ScheduleTimer("FindAndUpdateCorpse", 1)
		CorpseFound = false
		return
	end

	CorpseFound = true

--	self:print("Continent: " .. CorpseContinent)

	--if CorpseContinent ~= -1 then -- We've found a continent they're on

		self:print("My continent: " .. CorpseContinent)

		x, y = 0, 0 -- nil out continent position

		if CorpseZoneId ~= -1 then -- We have a cached zone from last check
			
			self:print("Testing old zone: " .. CorpseZoneId)

			SetMapZoom(CorpseContinent, CorpseZoneId)
			x, y = GetCorpseMapPosition()
		end

		if Pcontinent and (Pcontinent == CorpseContinent) and (x == 0 and y == 0) and (Pzone ~= 0) then

			CorpseZoneId = Pzone
			self:print("Testing current zone: " .. CorpseZoneId)

			SetMapZoom(CorpseContinent, CorpseZoneId)
			x, y = GetCorpseMapPosition()

		end

		--self:print("Zone: " .. CorpseZoneId)
		--self:print("X: " .. x .. "  Y: " .. y);

		if x == 0 and y == 0 then -- They're no longer in that zone or we never had a zone
			
			self:print("My Zone: " .. CorpseContinent)

			for i=1,select("#", GetMapZones(CorpseContinent)) do
				SetMapZoom(CorpseContinent, i)
				x, y = GetCorpseMapPosition()

				if x > 0 and y > 0 then -- Found them! Store the info
					CorpseZoneId = i
					CorpseX = x
					CorpseY = y
					CorpseCoords = HandyNotes:getCoord(x, y)

					local CorpseZone = select(CorpseZoneId, GetMapZones(CorpseContinent))
					self:print("Found corpse: " .. CorpseZone .. " [" .. x .. ":" .. y .. "]")
					--self:AddCharonCorpseNote(CorpseContinent, CorpseZoneId, CorpseZone, x, y)
					break
				end
			end
		else
			CorpseX = x
			CorpseY = y
			CorpseCoords = HandyNotes:getCoord(x, y)

			local CorpseZone = select(CorpseZoneId, GetMapZones(CorpseContinent))
			self:print("Found corpse in old zone: " .. CorpseZone .. " [" .. x .. ":" .. y .. "]")
			--self:AddCharonCorpseNote(CorpseContinent, CorpseZoneId, CorpseZone, x, y)
		end
	--end
end

--[[
function HC:AddNPCNote(vType)
	local vName = UnitName("npc")
	local vGuild = FigureNPCGuild("npc")

	local vGuid = UnitGUID("npc")
--	local vNpcid = tonumber(vGuid:sub(-12, -7), 16)
	-- XXX 3.3 compat - solution inspired by MobSpells addon
	local vNpcid = QueryQuestsCompleted and tonumber(vGuid:sub(-12, -9), 16) or tonumber(vGuid:sub(-12, -7), 16)


	self:AddCharonNote(vNpcid, vName, vGuild, vType)

end

]]

function HC:AddSpiritHealerNote()

	local Scontinent = SpiritContinent
	local Szone = SpiritZoneId
	local Sx, Sy = SpiritX, SpiritY
	local Scoord = SpiritCoords

	local SInfo = "H"
	local Smap = HandyNotes:GetMapFile(Scontinent, Szone)

	if (Smap) then
		for Scoords, name in pairs(db.factionrealm.nodes[Smap]) do
			if SInfo == name then
				local cx, cy = HandyNotes:getXY(Scoords)
				local dist = Astrolabe:ComputeDistance(Scontinent, Szone, Sx, Sy, Scontinent, Szone, cx, cy)
				if dist <= thres then -- Note already exists here
					Scoord = Scoords  -- let's not move
					SpiritCoords = Scoords -- update SpiritHealer location from the old loc for CorpseNote
					self:print("SpiritHealerExists: " .. Smap .. " - " .. Scoord .. " : " .. name)
					return
				end
			end
		end
		-- no Spirit Healer found near this location, add new note
		db.factionrealm.nodes[Smap][Scoord] = SInfo

		self:print("SpiritHealerNote: " .. Smap .. "[" .. Scoord .. "]: " .. SInfo)
		self:SendMessage("HandyNotes_NotifyUpdate", "Charon")
	end
end

function HC:AddCorpseNote()
	local Ccontinent = CorpseContinent
	local Czone = CorpseZoneId
	local Cx, Cy = CorpseX, CorpseY
	local Ccoord = CorpseCoords

	local Scoord = "" -- not known yet

	local CInfo = LastDamageSourceType .. ":" .. time() .. ":" .. LastDamageSource .. ":" .. Scoord

	local Cmap = HandyNotes:GetMapFile(Ccontinent, Czone)

	if (Cmap) then
		db.char.nodes[Cmap][Ccoord] = CInfo

		self:print("CorpseNote: " .. Cmap .. "[" .. Ccoord .. "]: " .. CInfo)
		self:SendMessage("HandyNotes_NotifyUpdate", "Charon")
	end
end

function HC:AddOrUpdateCorpseNote()
	local Ccontinent = CorpseContinent
	local Czone = CorpseZoneId
	local Cx, Cy = CorpseX, CorpseY
	local Ccoord = CorpseCoords

	local Scontinent = SpiritContinent
	local Szone = SpiritZoneId
	local Sx, Sy = SpiritX, SpiritY
	local Scoord = SpiritCoords

	if Scontinent ~= Ccontinent then 
		self:print("Error: Spirit healer on different continent: '" .. (select(Scontinent, GetMapContinents()) or "nil") .. "' from corpse: '" .. (select(Ccontinent, GetMapContinents()) or "nil") .. "'");
		return
	end

	if Szone ~= Czone then
		local Tx, Ty = Astrolabe:TranslateWorldMapPosition(Scontinent, Szone, Sx, Sy, Ccontinent, Czone)
		if not Tx then 
			self:print("Error: Spirit healer coordinates translation failed!");
			return
		end
		Scoord = HandyNotes:getCoord(Tx, Ty)
	end

	local CInfo = nil

	local Cmap = HandyNotes:GetMapFile(Ccontinent, Czone)

	if (Cmap) then
		local oldInfo
		if db.char.nodes[Cmap] and db.char.nodes[Cmap][Ccoord] then
			oldInfo = db.char.nodes[Cmap][Ccoord]
		end
		if oldInfo then
			local vType, vTime, vMonster, vSPCoord = strsplit(":", oldInfo)
			-- TODO: check type, time, empty coords
			CInfo = vType .. ":" .. vTime .. ":" .. (vMonster or "").. ":" .. Scoord
		else
			CInfo = LastDamageSourceType .. ":" .. time() .. ":" .. LastDamageSource .. ":" .. Scoord
		end
		db.char.nodes[Cmap][Ccoord] = CInfo

		self:print("CorpseNoteU: " .. Cmap .. "[" .. Ccoord .. "]: " .. CInfo)
		self:SendMessage("HandyNotes_NotifyUpdate", "Charon")
	end

end

function HC:AddCharonCorpseNote(continent, zone, mapname, x, y)

-- get ghost position

	local Gcontinent, Gzone, Gx, Gy = Astrolabe:GetCurrentPlayerPosition()
--	if not Gcontinent then
--		return
--	end

	--local vInfo = vType .. ":" .. vName .. ":" .. vGuild .. ":" .. vNpcid

	local coord = HandyNotes:getCoord(x, y)
	local map = HandyNotes:GetMapFile(continent, zone)

	local Gcoord = HandyNotes:getCoord(Gx, Gy)
	local Gmap = HandyNotes:GetMapFile(Gcontinent, Gzone)

	if mapname ~= map then 
		self:print("Warning: mapname is different: '" .. map .. "' for zone: '" .. mapname .. "'");
	end

	if Gmap ~= map then 
		self:print("Warning: Ghost in different zone: '" .. Gmap .. "' from corpse: '" .. map .. "'");
	end

	local GInfo = "S";

	if map then
		for Gcoords, name in pairs(db.factionrealm.nodes[map]) do
			if GInfo == name then
				local cx, cy = HandyNotes:getXY(Gcoords)
				local dist = Astrolabe:ComputeDistance(Gcontinent, Gzone, Gx, Gy, Gcontinent, Gzone, cx, cy)
				if dist <= thres then -- Note already exists here
					Gcoord = Gcoords  -- let's not move, just update vInfo
--					return
--				else -- Note exists on different location = has moved -> delete old info
--					db.factionrealm.nodes[map][coords] = nil
				end
			end
		end
		db.factionrealm.nodes[map][Gcoord] = GInfo

		local vInfo = "D:" .. time() .. ":" .. ":" .. Gcoord;
		db.factionrealm.nodes[map][coord] = vInfo

		self:SendMessage("HandyNotes_NotifyUpdate", "Charon")
	end

	-- now create spirit healer note
	local continent, zone, x, y = Astrolabe:GetCurrentPlayerPosition()
	if not vName or not continent then
		return
	end
end

------------------------------------------------------------------------------------------------------
-- The following function is from Daniel Stephens <iriel@vigilance-committee.org>
-- with reference to TaxiFrame.lua in Blizzard's UI and Graph-1.0 Ace2 library (by Cryect) 
local TAXIROUTE_LINEFACTOR = 128/126; -- Multiplying factor for texture coordinates
local TAXIROUTE_LINEFACTOR_2 = TAXIROUTE_LINEFACTOR / 2; -- Half of that

-- T        - Texture
-- C        - Canvas Frame (for anchoring)
-- sx,sy    - Coordinate of start of line
-- ex,ey    - Coordinate of end of line
-- w        - Width of line
-- relPoint - Relative point on canvas to interpret coords (Default BOTTOMLEFT)
function G:DrawLine(C, sx, sy, ex, ey, w, color, layer)
	local relPoint = "BOTTOMLEFT"

	if not C.HandyNotesCh_Lines then
		C.HandyNotesCh_Lines = {}
		C.HandyNotesCh_Lines_Used = {}
	end

	local T = tremove(C.HandyNotesCh_Lines) or C:CreateTexture(nil, "ARTWORK")
	T:SetTexture("Interface\\AddOns\\HandyNotes_Charon\\line")
	tinsert(C.HandyNotesCh_Lines_Used,T)

	T:SetDrawLayer(layer or "ARTWORK")

	T:SetVertexColor(color[1],color[2],color[3],color[4]);
	-- Determine dimensions and center point of line
	local dx,dy = ex - sx, ey - sy;
	local cx,cy = (sx + ex) / 2, (sy + ey) / 2;

	-- Normalize direction if necessary
	if (dx < 0) then
		dx,dy = -dx,-dy;
	end

	-- Calculate actual length of line
	local l = ((dx * dx) + (dy * dy)) ^ 0.5;

	-- Sin and Cosine of rotation, and combination (for later)
	local s,c = -dy / l, dx / l;
	local sc = s * c;

	-- Calculate bounding box size and texture coordinates
	local Bwid, Bhgt, BLx, BLy, TLx, TLy, TRx, TRy, BRx, BRy;
	if (dy >= 0) then
		Bwid = ((l * c) - (w * s)) * TAXIROUTE_LINEFACTOR_2;
		Bhgt = ((w * c) - (l * s)) * TAXIROUTE_LINEFACTOR_2;
		BLx, BLy, BRy = (w / l) * sc, s * s, (l / w) * sc;
		BRx, TLx, TLy, TRx = 1 - BLy, BLy, 1 - BRy, 1 - BLx;
		TRy = BRx;
	else
		Bwid = ((l * c) + (w * s)) * TAXIROUTE_LINEFACTOR_2;
		Bhgt = ((w * c) + (l * s)) * TAXIROUTE_LINEFACTOR_2;
		BLx, BLy, BRx = s * s, -(l / w) * sc, 1 + (w / l) * sc;
		BRy, TLx, TLy, TRy = BLx, 1 - BRx, 1 - BLx, 1 - BLy;
		TRx = TLy;
	end

	-- Thanks Blizzard for adding (-)10000 as a hard-cap and throwing errors!
	-- The cap was added in 3.1.0 and I think it was upped in 3.1.1
	--  (way less chance to get the error)
	if TLx > 10000 then TLx = 10000 elseif TLx < -10000 then TLx = -10000 end
	if TLy > 10000 then TLy = 10000 elseif TLy < -10000 then TLy = -10000 end
	if BLx > 10000 then BLx = 10000 elseif BLx < -10000 then BLx = -10000 end
	if BLy > 10000 then BLy = 10000 elseif BLy < -10000 then BLy = -10000 end
	if TRx > 10000 then TRx = 10000 elseif TRx < -10000 then TRx = -10000 end
	if TRy > 10000 then TRy = 10000 elseif TRy < -10000 then TRy = -10000 end
	if BRx > 10000 then BRx = 10000 elseif BRx < -10000 then BRx = -10000 end
	if BRy > 10000 then BRy = 10000 elseif BRy < -10000 then BRy = -10000 end

	-- Set texture coordinates and anchors
	T:ClearAllPoints();
	T:SetTexCoord(TLx, TLy, BLx, BLy, TRx, TRy, BRx, BRy);
	T:SetPoint("BOTTOMLEFT", C, relPoint, cx - Bwid, cy - Bhgt);
	T:SetPoint("TOPRIGHT",   C, relPoint, cx + Bwid, cy + Bhgt);
	T:Show()
	return T
end

function G:HideLines(C)
	if C.HandyNotesCh_Lines then
		for i = #C.HandyNotesCh_Lines_Used, 1, -1 do
			C.HandyNotesCh_Lines_Used[i]:Hide()
			tinsert(C.HandyNotesCh_Lines, tremove(C.HandyNotesCh_Lines_Used))
		end
	end
end